@@ -23,7 +23,7 @@
 ;;; In:                 Stream, Eof-Errorp, Eof-Value
 ;;; Bin:                Stream, Eof-Errorp, Eof-Value
 ;;; N-Bin:              Stream, Buffer, Start, Numbytes, Eof-Errorp
-;;; Out:                Stream, Character
+;;; Cout:               Stream, Character
 ;;; Bout:               Stream, Integer
 ;;; Sout:               Stream, String, Start, End
 ;;; Misc:               Stream, Operation, &Optional Arg1, Arg2
@@ -49,7 +49,7 @@
 ;;;  :line-length       - Return the length of a line of output.
 ;;;  :charpos           - Return current output position on the line.
 ;;;  :file-length       - Return the file length of a file stream.
-;;;  :file-position     - Return or change the current position of a
+;;;  :[gs]et-file-position  - Return or change the current position of a
 ;;;                       file stream.
 ;;;  :file-name         - Return the name of an associated file.
 ;;;  :interactive-p     - Is this an interactive device?
@@ -90,6 +90,31 @@
 (deftype ansi-stream-cin-buffer ()
   `(simple-array character (,+ansi-stream-in-buffer-length+)))
 
+(eval-when (:compile-toplevel :load-toplevel :execute)
+(defun %stream-opcode (name)
+  (ecase name
+    (:listen              0)
+    (:unread              1)
+    (:force-output        2)
+    (:finish-output       3)
+    (:charpos             4)
+    (:get-file-position   5)
+    (:set-file-position   6)
+    (:element-type        7)
+    (:element-mode        8)
+    (:external-format     9)
+    (:line-length        10)
+    (:file-length        11)
+    (:file-string-length 12)
+    (:clear-input        13)
+    (:clear-output       14)
+    (:close              15)
+    (:interactive-p      16)
+    ;; This is used by string streams and is not an opcode seen
+    ;; by STREAM-MISC-DISPATCH. The new stream pseudomethod convention
+    ;; can't pass random keywords to the misc method.
+    (:reset-unicode-p    17))))
+
 ;;; base class for ANSI standard streams (as opposed to the Gray
 ;;; streams extension)
 (defstruct (ansi-stream (:constructor nil)
@@ -105,26 +130,29 @@
   (in-index +ansi-stream-in-buffer-length+
             :type (integer 0 #.+ansi-stream-in-buffer-length+))
 
+  ;; FIXME: declare all these function types more rigorously.
+
   ;; buffered input functions
   (in #'ill-in :type function)                  ; READ-CHAR function
   (bin #'ill-bin :type function)                ; byte input function
   ;; 'n-bin' might not transfer bytes to the consumer.
   ;; A character FD-STREAM uses this method to transfer octets from the
   ;; source buffer into characters of the destination buffer.
-  (n-bin #'ill-bin :type function)              ; n-byte input function
+  (n-bin #'ill-bin :type                        ; n-byte input function
+   (sfunction (stream (simple-unboxed-array (*)) index index t) index))
 
   ;; output functions
-  (out #'ill-out :type function)                ; WRITE-CHAR function
+  (cout #'ill-out :type function)               ; WRITE-CHAR function
   (bout #'ill-bout :type function)              ; byte output function
   (sout #'ill-out :type function)               ; string output function
 
   ;; other, less-used methods
-  (misc #'no-op-placeholder :type function)
+  (misc #'no-op-placeholder :type (function (stream (integer 0 17) t) *))
 
   ;; Absolute character position, acting also as a generalized boolean
   ;; in lieu of testing FORM-TRACKING-STREAM-P to see if we must
   ;; maintain correctness of the slot in ANSI-STREAM-UNREAD-CHAR.
-  (input-char-pos nil))
+  (input-char-pos nil :type (or null index)))
 
 ;;; SYNONYM-STREAM type is needed by ANSI-STREAM-{INPUT,OUTPUT}-STREAM-P
 ;;; and also needed by OPEN (though not obviously), which is compiled
@@ -134,7 +162,7 @@
                                      (in #'synonym-in)
                                      (bin #'synonym-bin)
                                      (n-bin #'synonym-n-bin)
-                                     (out #'synonym-out)
+                                     (cout #'synonym-out)
                                      (bout #'synonym-bout)
                                      (sout #'synonym-sout)
                                      (misc #'synonym-misc))
@@ -144,6 +172,22 @@
   (symbol nil :type symbol :read-only t))
 (declaim (freeze-type synonym-stream))
 
+(defstruct (broadcast-stream (:include ansi-stream
+                                       (cout #'broadcast-cout)
+                                       (bout #'broadcast-bout)
+                                       (sout #'broadcast-sout)
+                                       (misc #'broadcast-misc))
+                             (:constructor %make-broadcast-stream
+                                           (streams))
+                             (:copier nil)
+                             (:predicate nil))
+  ;; a list of all the streams we broadcast to
+  (streams () :type list :read-only t))
+(declaim (freeze-type broadcast-stream))
+
+(define-load-time-global *null-broadcast-stream* (make-broadcast-stream))
+(declaim (type stream *null-broadcast-stream*))
+
 (defmethod print-object ((x stream) stream)
   (print-unreadable-object (x stream :type t :identity t)))
 
@@ -212,8 +256,9 @@
                  ;;  the variable var." - so we can read it here with impunity.
                  `(unwind-protect (progn ,@forms) (close ,var))
                  `(progn ,@forms)))) ; don't bother closing
-      `(,bind ((,dummy (%make-instance ,(dd-length (find-defstruct-description
-                                                    'string-input-stream)))))
+      `(,bind ((,dummy (%make-structure-instance
+                        ,(find-defstruct-description 'string-input-stream)
+                        nil)))
               ,(multiple-value-bind (forms decls) (parse-body forms-decls nil)
                  (if index
                      `(multiple-value-bind (,var ,offset) ,ctor
@@ -255,15 +300,15 @@
   ;; end of the stream.
   (index-cache 0 :type index)
   ;; Pseudo-actual element type. We no longer store the as-requested type.
-  ;; (If the value is :DEFAULT, we return CHARACTER on inquiry.)
+  ;; (If the value is *, we return CHARACTER on inquiry.)
   (element-type nil :read-only t
-                    :type (member #+sb-unicode :default
+                    :type (member #+sb-unicode *
                                   #+sb-unicode character
                                   base-char nil)))
 (declaim (freeze-type string-output-stream))
 
 (defmacro %allocate-string-ostream ()
-  `(%make-instance ,(dd-length (find-defstruct-description 'string-output-stream))))
+  `(%make-structure-instance ,(find-defstruct-description 'string-output-stream) nil))
 
 (defmacro with-output-to-string
     ((var &optional string &key (element-type ''character)) &body body)
@@ -272,8 +317,9 @@
         ;; "If string is supplied, element-type is ignored".
         ;; Why do implementors take this to mean "evaluated and ignored?"
         ;; I would have figured it meant expansion-time ignored.
-        `(dx-let ((,dummy (%make-instance ,(dd-length (find-defstruct-description
-                                                       'fill-pointer-output-stream)))))
+        `(dx-let ((,dummy (%make-structure-instance
+                           ,(find-defstruct-description 'fill-pointer-output-stream)
+                           nil)))
            (let ((,var (truly-the fill-pointer-output-stream
                                   (%init-fill-pointer-output-stream
                                    ,dummy ,string ,element-type))))
@@ -289,3 +335,89 @@
 ;;; This is not something that the standard macro permits.
 (defmacro %with-output-to-string ((var) &body body)
   (expand-with-output-to-string var ''character body t))
+
+;;; A macro to better exploit jump tables. Even though jumping based on symbol-hash
+;;; is possible, it is of course slightly faster to dispatch on small integers,
+;;; and for architectures which don't implement jump tables,
+;;; using integers eliminates the load of many code header constants.
+;;; Streams which don't want to handle every operation (don't end in a T clause)
+;;; should specify :DEFAULT NIL to avoid an error.
+(defmacro stream-misc-case ((operation &key (default 'error)) &rest clauses)
+  (let* ((otherwise)
+         (clauses
+          (mapcar (lambda (clause)
+                    (let ((key (car clause)))
+                      (cons (if (eq key 't)
+                                (setq otherwise t)
+                                (mapcar #'%stream-opcode (ensure-list (car clause))))
+                            (cdr clause))))
+                  clauses)))
+    `(,(if (and (not otherwise) (eq default 'error)) 'ecase 'case) ,operation
+      ,@clauses)))
+
+(defmacro call-ansi-stream-misc (stream operation &optional (arg nil argp))
+  ;; Never stuff in a placeholder in the three operations that take an argument.
+  (aver (or argp (not (memq operation '(:set-file-position :file-string-length :unread)))))
+  `(funcall (ansi-stream-misc ,stream) ,stream
+            ;; If operation is a literal keyword, translate it, otherwise
+            ;; it is a variable whose values is the opcode for passthru.
+            ,(if (keywordp operation) (%stream-opcode operation) operation)
+            ,(if argp arg 0)))
+
+;;; This predicate assumes some things:
+;;;  - that a simple-stream class will never be FUNCALLABLE-STANDARD-OBJECT
+;;;    (so that we need only check for %INSTANCEP)
+;;;  - a class will never be redefined such that it moves from one part of the stream
+;;;    hierarchy to the other. (i.e. you don't redefine MY-STREAM-CLASS such that
+;;;    it was a Gray stream and becomes a simple-stream or vice-versa)
+;;;  - instance invalidation, if it needs to happen, will happen somewhere later in
+;;;    the stream protocol. (the layout-invalid trap can be deferred)
+;;;    This is OK because the next level of dispatch is not to a generic function
+;;;    in the simple-stream layering. It is generic at the "device" but not the API.
+(defmacro simple-stream-p (x)
+  `(and (%instancep ,x)
+        (logtest (layout-flags (%instance-layout ,x))
+                 sb-kernel::+simple-stream-layout-flag+)))
+
+;;; This macro is for the first-level of dispatch which usually entails
+;;; deciding whether the argument is actually a stream or merely a designator
+;;; for a stream (T or NIL), and then figuring out which family of stream
+;;; it belongs to: ANSI, Gray, or simple.
+(defmacro stream-api-dispatch ((streamvar &optional designator) &key native simple gray)
+  ;; Most CL: stream APIs use explicit-check, so we should assert that the thing
+  ;; is actually a stream.
+  ;; If the CL interface bypasses STREAM-API-DISPATCH then it is up to generic layer
+  ;; to signal an error, the class of which seems unfortunately subject to debate.
+  (aver (and native (or simple gray)))
+  ;; Dispatch native first, then if sb-simple-streams have been loaded,
+  ;; those, and finally punt to a generic function.
+  `(block stream
+     (tagbody
+      again
+        ;; Assume that simple-stream can not inherit funcallable-standard-object.
+        (when (%instancep ,streamvar)
+          (let ((layout (%instance-layout ,streamvar)))
+            (cond ((sb-c::%structure-is-a layout ,(find-layout 'ansi-stream))
+                   (let ((,streamvar (truly-the ansi-stream ,streamvar)))
+                     (return-from stream ,native)))
+                  ((logtest (layout-flags layout) +simple-stream-layout-flag+)
+                   (return-from stream ,simple)))))
+        (cond ((streamp ,streamvar)
+               (return-from stream ,gray))
+              ,@(case designator
+                  (:input
+                   `(((eq ,streamvar t)
+                       (setf ,streamvar *terminal-io*)
+                      (go again))
+                     ((null ,streamvar)
+                      (setf ,streamvar *standard-input*)
+                      (go again))))
+                   (:output
+                    `(((eq ,streamvar t)
+                       (setf ,streamvar *terminal-io*)
+                       (go again))
+                      ((null ,streamvar)
+                       (setf ,streamvar *standard-output*)
+                       (go again)))))
+              (t
+               (sb-c::%type-check-error/c ,streamvar 'sb-kernel::object-not-stream-error nil))))))
