@@ -17,6 +17,7 @@
 #define _GNU_SOURCE
 
 #include <stdio.h>
+#include <stdbool.h>
 #include <signal.h>
 #include "sbcl.h"
 #include "runtime.h"
@@ -28,6 +29,7 @@
 #include "arch.h"
 #include "genesis/compiled-debug-fun.h"
 #include "genesis/compiled-debug-info.h"
+#include "genesis/hash-table.h"
 #include "genesis/package.h"
 #include "genesis/static-symbols.h"
 #include "genesis/primitive-objects.h"
@@ -35,17 +37,16 @@
 #include "gc.h"
 #include "code.h"
 #include "var-io.h"
-#include "gc-internal.h"
-#include "forwarding-ptr.h"
 #include "lispstring.h"
 
 #ifdef LISP_FEATURE_OS_PROVIDES_DLADDR
 # include <dlfcn.h>
 #endif
 
-static int decode_locs(lispobj packed_integer, int *offset, int *elsewhere)
+int df_decode_locs(lispobj encoded, int *offset, int *elsewhere)
 {
     struct varint_unpacker unpacker;
+    lispobj packed_integer = listp(encoded) ? CONS(encoded)->cdr : encoded;
     varint_unpacker_init(&unpacker, packed_integer);
     return varint_unpack(&unpacker, offset) && varint_unpack(&unpacker, elsewhere);
 }
@@ -67,14 +68,14 @@ debug_function_from_pc (struct code* code, void *pc)
 
     struct compiled_debug_fun *df = (struct compiled_debug_fun*)native_pointer(di->fun_map);
     int begin, end, elsewhere_begin, elsewhere_end;
-    if (!decode_locs(df->encoded_locs, &begin, &elsewhere_begin))
+    if (!df_decode_locs(df->encoded_locs, &begin, &elsewhere_begin))
         return NULL;
     sword_t offset = (char*)pc - code_text_start(code);
     while (df) {
         struct compiled_debug_fun *next;
         if (df->next != NIL) {
             next = (struct compiled_debug_fun*) native_pointer(df->next);
-            if (!decode_locs(next->encoded_locs, &end, &elsewhere_end))
+            if (!df_decode_locs(next->encoded_locs, &end, &elsewhere_end))
                 return NULL;
         } else {
             next = 0;
@@ -99,7 +100,7 @@ print_string (struct vector *vector, FILE *f)
         return;
     }
     int i;
-    int n = fixnum_value(vector->length);
+    int n = vector_len(vector);
     for (i = 0; i < n; i++) {
         unsigned int c = schar(vector, i);
         if (c > 0xFFFF) fprintf(f,"\\U%08x", c);
@@ -118,14 +119,27 @@ lispobj debug_print(lispobj string)
 {
     print_string(VECTOR(string), stderr);
     putc('\n', stderr);
+    fflush(stderr);
     return 0;
 }
 
-static int string_equal (struct vector *vector, char *string)
+lispobj symbol_package(struct symbol* s)
 {
-    if (widetag_of(&vector->header) != SIMPLE_BASE_STRING_WIDETAG)
-        return 0;
-    return !strcmp((char *) vector->data, string);
+    static int warned;
+    // If using ldb when debugging cold-init, this can be confusing to see all symbols
+    // as if they were uninterned, but package-IDs are always available in the symbol.
+    // End-users should never see this failure.
+    if (!lisp_package_vector) {
+        if (!warned) {
+          fprintf(stderr, "Warning: lisp package array is not initialized for C\n");
+          warned = 1;
+        }
+        return NIL;
+    }
+    struct vector* v = VECTOR(lisp_package_vector);
+    int id = symbol_package_id(s);
+    if (id < vector_len(v)) return v->data[id];
+    lose("can't decode package ID %d", id);
 }
 
 static void
@@ -148,26 +162,26 @@ print_entry_name (lispobj name, FILE *f)
         putc(')', f);
     } else if (lowtag_of(name) == OTHER_POINTER_LOWTAG) {
         struct symbol *symbol = SYMBOL(name);
+        char* prefix = 0;
         int widetag = header_widetag(symbol->header);
         switch (widetag) {
         case SYMBOL_WIDETAG:
-            if (symbol->package != NIL) {
-                struct package *pkg
-                    = (struct package *) native_pointer(follow_maybe_fp(symbol->package));
-                struct vector *pkg_name = VECTOR(follow_maybe_fp(pkg->_name));
-                if (string_equal(pkg_name, "COMMON-LISP"))
-                    ;
-                else if (string_equal(pkg_name, "COMMON-LISP-USER")) {
-                    fputs("CL-USER::", f);
-                }
-                else if (string_equal(pkg_name, "KEYWORD")) {
-                    putc(':', f);
-                } else {
-                    print_string(pkg_name, f);
-                    fputs("::", f);
-                }
+            switch (symbol_package_id(symbol)) {
+            case PACKAGE_ID_NONE: prefix = "#:"; break;
+            case PACKAGE_ID_LISP: prefix = ""; break;
+            case PACKAGE_ID_USER: prefix = "CL-USER::"; break;
+            case PACKAGE_ID_KEYWORD: prefix = ":"; break;
             }
-            print_string(VECTOR(follow_maybe_fp(symbol->name)), f);
+            if (prefix) fputs(prefix, f); else {
+                struct package *pkg
+                    = (struct package *)native_pointer(symbol_package(symbol));
+                lispobj name_ptr = follow_maybe_fp(pkg->_name);
+                if (name_ptr) {
+                    print_string(VECTOR(name_ptr), f);
+                }
+                fputs("::", f);
+            }
+            print_string(symbol_name(symbol), f);
             break;
         case SIMPLE_BASE_STRING_WIDETAG:
 #ifdef SIMPLE_CHARACTER_STRING_WIDETAG
@@ -187,7 +201,7 @@ print_entry_name (lispobj name, FILE *f)
     }
 }
 
-static void
+static void __attribute__((unused))
 print_entry_points (struct code *code, FILE *f)
 {
     int n_funs = code_n_funs(code);
@@ -207,26 +221,7 @@ print_entry_points (struct code *code, FILE *f)
 /* KLUDGE: Sigh ... I know what the call frame looks like and it had
  * better not change. */
 
-struct call_frame {
-        struct call_frame *old_cont;
-        lispobj saved_lra;
-        lispobj code;
-        lispobj other_state[5];
-};
-
-struct call_info {
-    struct call_frame *frame;
-    int interrupted;
-    struct code *code;
-    lispobj lra;
-    int pc; /* Note: this is the trace file offset, not the actual pc. */
-};
-
-// simple-fun headers have a pointer to layout-of-function in the
-// upper bytes if words are 8 bytes, so mask off those bytes.
-#define HEADER_LENGTH(header) (((header)>>8) & FUN_HEADER_NWORDS_MASK)
-
-static int previous_info(struct call_info *info);
+#include "callframe.inc"
 
 static struct code *
 code_pointer(lispobj object)
@@ -236,9 +231,11 @@ code_pointer(lispobj object)
     switch (widetag_of(headerp)) {
         case CODE_HEADER_WIDETAG:
             break;
+#ifdef RETURN_PC_WIDETAG
         case RETURN_PC_WIDETAG:
+#endif
         case SIMPLE_FUN_WIDETAG:
-            len = HEADER_LENGTH(*headerp);
+            len = (HeaderValue(*headerp) & FUN_HEADER_NWORDS_MASK);
             if (len == 0)
                 headerp = NULL;
             else
@@ -251,26 +248,13 @@ code_pointer(lispobj object)
     return (struct code *) headerp;
 }
 
-static boolean
-cs_valid_pointer_p(struct call_frame *pointer)
+static bool
+cs_valid_pointer_p(struct thread *thread, struct call_frame *pointer)
 {
-    struct thread *thread=arch_os_get_current_thread();
     return (((char *) thread->control_stack_start <= (char *) pointer) &&
             ((char *) pointer < (char *) access_control_stack_pointer(thread)));
 }
 
-static void
-call_info_from_lisp_state(struct call_info *info)
-{
-    info->frame = (struct call_frame *)access_control_frame_pointer(arch_os_get_current_thread());
-    info->interrupted = 0;
-    info->code = NULL;
-    info->lra = 0;
-    info->pc = 0;
-
-    previous_info(info);
-}
-
 static void
 call_info_from_context(struct call_info *info, os_context_t *context)
 {
@@ -284,39 +268,44 @@ call_info_from_context(struct call_info *info, os_context_t *context)
         info->frame =
             (struct call_frame *)(uword_t)
                 (*os_context_register_addr(context, reg_OCFP));
+#ifdef reg_LRA
         info->lra = (lispobj)(*os_context_register_addr(context, reg_LRA));
+#else
+        info->lra = (lispobj)(*os_context_register_addr(context, reg_RA));
+#endif
         info->code = code_pointer(info->lra);
         pc = (uword_t)native_pointer(info->lra);
     } else
 #endif
     {
+        pc = os_context_pc(context);
         info->frame =
             (struct call_frame *)(uword_t)
                 (*os_context_register_addr(context, reg_CFP));
         info->code =
+#ifdef reg_CODE
             code_pointer(*os_context_register_addr(context, reg_CODE));
+#else
+        (struct code *)component_ptr_from_pc((char *)pc);
+#endif
         info->lra = NIL;
-        pc = *os_context_pc_addr(context);
+
     }
+
     if (info->code != NULL)
-        info->pc = pc - (uword_t) info->code -
-            (HEADER_LENGTH(info->code->header) * sizeof(lispobj));
+        info->pc = (char*)pc - (char*)info->code;
     else
         info->pc = 0;
 }
 
-static int
-previous_info(struct call_info *info)
+// Return 1 if we have a valid frame, 0 if not.
+int lisp_frame_previous(struct thread *thread, struct call_info *info)
 {
     struct call_frame *this_frame;
-    struct thread *thread=arch_os_get_current_thread();
     int free_ici;
     lispobj lra;
 
-    if (!cs_valid_pointer_p(info->frame)) {
-        printf("Bogus callee value (0x%lx).\n", (long)info->frame);
-        return 0;
-    }
+    if (!cs_valid_pointer_p(thread, info->frame)) return 0;
 
     this_frame = info->frame;
     info->lra = this_frame->saved_lra;
@@ -339,15 +328,22 @@ previous_info(struct call_info *info)
             }
         }
     } else if (fixnump(lra)) {
-        info->code = (struct code*)native_pointer(this_frame->code);
-        // FIXME: is this right? fixnumish LRAs are based off the object base address
-        // and not the code text start?
-        info->pc = (uword_t)(info->code + lra);
+        info->code =
+#ifdef reg_CODE
+        (struct code*)native_pointer(this_frame->code);
+#else
+        (struct code*)component_ptr_from_pc((char *)lra);
+#endif
+#ifdef reg_LRA
+        info->pc = lra;
+#else
+        info->pc = (char*)native_pointer(lra) - (char*)info->code;
+#endif
         info->lra = NIL;
     } else {
         info->code = code_pointer(lra);
         if (info->code != NULL)
-            info->pc = (char*)native_pointer(info->lra) - code_text_start(info->code);
+            info->pc = (char*)native_pointer(info->lra) - (char*)info->code;
         else
             info->pc = 0;
     }
@@ -358,46 +354,91 @@ previous_info(struct call_info *info)
 void
 lisp_backtrace(int nframes)
 {
+    struct thread *thread = get_sb_vm_thread();
     struct call_info info;
+
+    info.frame = (struct call_frame *)access_control_frame_pointer(thread);
+    info.interrupted = 0;
+    info.code = NULL;
+    info.lra = 0;
+    info.pc = 0;
+
     int i = 0;
-    call_info_from_lisp_state(&info);
-
+    int footnotes = 0;
     do {
+        if (!lisp_frame_previous(thread, &info)) {
+            if (info.frame) // 0 is normal termination of the call chain
+                printf("Bad frame pointer %p [valid range=%p..%p]\n", info.frame,
+                       thread->control_stack_start, thread->control_stack_end);
+            break;
+        }
         printf("%4d: ", i);
+        // Print spaces to keep the alignment nice
+        if (info.interrupted
+#ifdef reg_LRA
+            || info.lra == NIL
+#endif
+            ) {
+            putchar('[');
+            if (info.interrupted) { footnotes |= 1; putchar('I'); }
+#ifdef reg_LRA
+            if (info.lra == NIL) { footnotes |= 2; putchar('*'); }
+#endif
+            putchar(']');
+            if (!(info.lra == NIL && info.interrupted)) putchar(' ');
+        } else {
+            printf("    ");
+        }
+        printf("%p ", info.frame);
+        void* absolute_pc = 0;
+        if (info.code) {
+            absolute_pc = (char*)info.code + info.pc;
+            printf("pc=%p {%p+%04x} ", absolute_pc, info.code, (int)info.pc);
+        } else {
+            absolute_pc = (char*)info.pc;
+            printf("pc=%p ", absolute_pc);
+        }
 
-        if (info.code != (struct code *) 0) {
-            struct compiled_debug_fun *df ;
-            if (info.lra != NIL &&
-                (df = debug_function_from_pc((struct code *)info.code, (void *)info.lra)))
+        // If LRA does not match the PC, print it. This should not happen.
+        if (info.lra != make_lispobj(absolute_pc, OTHER_POINTER_LOWTAG)
+            && info.lra != NIL)
+            printf("LRA=%p ", (void*)info.lra);
+
+        int fpvalid = (lispobj*)info.frame >= thread->control_stack_start
+          && (lispobj*)info.frame < thread->control_stack_end;
+
+        // If the FP is invalid, then quite likely we'd crash trying to find a
+        // compiled-debug-fun because info.code is a wild pointer
+        if (!fpvalid) { printf(" BAD FRAME\n"); break; }
+
+        if (info.code) {
+            struct compiled_debug_fun *df;
+            if (absolute_pc &&
+                (df = debug_function_from_pc((struct code *)info.code, absolute_pc)))
                 print_entry_name(df->name, stdout);
             else
-                print_entry_points((struct code *)info.code, stdout);
-
-            printf(" %p", (void*)((uword_t) info.code | OTHER_POINTER_LOWTAG));
+                // I can't imagine a scenario where we have info.code
+                // but do not have an absolute_pc, or debug-fun can't be found.
+                // Anyway, we can uniquely identify code by serial# now.
+                printf("{code_serialno=%x}", code_serialno(info.code));
         }
-        else
-            printf("CODE = ???");
-        printf("%s fp = %p", info.interrupted ? " [interrupted]" : "",
-               info.frame);
 
-        if (info.lra != NIL)
-            printf(" LRA = %p", (void*)info.lra);
-        else
-            printf(" <no LRA>");
-
-        if (info.pc)
-            printf(" pc_ofs = %p", (void*)(long)info.pc);
         putchar('\n');
 
-    } while (i++ < nframes && previous_info(&info));
+    } while (++i <= nframes);
+    if (footnotes) printf("Note: [I] = interrupted"
+#ifdef reg_LRA
+                          ", [*] = no LRA"
+#endif
+                          "\n");
 }
 
 #else
 
 static int
-altstack_pointer_p (void __attribute__((unused)) *p) {
+altstack_pointer_p(__attribute__((unused)) struct thread* thread,
+                   __attribute__((unused)) void *p) {
 #ifndef LISP_FEATURE_WIN32
-    struct thread* thread = arch_os_get_current_thread();
     // FIXME: shouldn't this be testing '>=' start and '<' end ?
     //        i.e. Was it only right because the calculations themselves were wrong ?
     return (p > calc_altstack_base(thread) && p <= calc_altstack_end(thread));
@@ -408,19 +449,18 @@ altstack_pointer_p (void __attribute__((unused)) *p) {
 }
 
 static int
-stack_pointer_p (void *p)
+stack_pointer_p(struct thread* thread, void *p)
 {
     /* we are using sizeof(long) here, because that is the right value on both
      * x86 and x86-64.  (But note that false positives would not cause much harm
      * given the heuristical nature of x86_call_context.) */
     uword_t stack_alignment = sizeof(void*);
     void *stack_start;
-    struct thread *thread = arch_os_get_current_thread();
 
-    if (altstack_pointer_p(p))
+    if (altstack_pointer_p(thread, p))
         return 1;
 
-    if (altstack_pointer_p(&p)) {
+    if (altstack_pointer_p(thread, &p)) {
         stack_start = (void *) thread->control_stack_start;
     } else {
         /* Use the current frame address, since there should be no
@@ -433,30 +473,28 @@ stack_pointer_p (void *p)
 }
 
 static int
-ra_pointer_p (void *ra)
+ra_pointer_p (struct thread* th, void *ra)
 {
-  /* the check against 4096 is still a mystery to everyone interviewed about
-   * it, but recent changes to sb-sprof seem to suggest that such values
-   * do occur sometimes. */
-  return ((uword_t) ra) > 4096 && !stack_pointer_p (ra);
+  return !stack_pointer_p (th, ra);
 }
 
 static int NO_SANITIZE_MEMORY
-x86_call_context (void *fp, void **ra, void **ocfp)
+x86_call_context (struct thread* th, void *fp, void **ra, void **ocfp)
 {
   void *c_ocfp;
   void *c_ra;
   int c_valid_p;
 
-  if (!stack_pointer_p(fp))
+  if (!stack_pointer_p(th, fp))
     return 0;
 
   c_ocfp    = *((void **) fp);
   c_ra      = *((void **) fp + 1);
 
+  // frame is valid even if the return address is bogus
   c_valid_p = (c_ocfp > fp
-               && stack_pointer_p(c_ocfp)
-               && ra_pointer_p(c_ra));
+               && stack_pointer_p(th, c_ocfp)
+               && ra_pointer_p(th, c_ra));
 
   if (c_valid_p)
     *ra = c_ra, *ocfp = c_ocfp;
@@ -469,7 +507,7 @@ x86_call_context (void *fp, void **ra, void **ocfp)
 void
 describe_thread_state(void)
 {
-    struct thread *thread = arch_os_get_current_thread();
+    struct thread *thread = get_sb_vm_thread();
     struct interrupt_data *data = &thread_interrupt_data(thread);
 #ifndef LISP_FEATURE_WIN32
     sigset_t mask;
@@ -479,37 +517,79 @@ describe_thread_state(void)
     if (string[0]) printf("Signal mask: %s\n", string);
 #endif
     printf("Specials:\n");
-    printf(" *GC-INHIBIT* = %s\n", (read_TLS(GC_INHIBIT, thread) == T) ? "T" : "NIL");
-    printf(" *GC-PENDING* = %s\n", (read_TLS(GC_PENDING, thread) == T) ? "T" : "NIL");
-    printf(" *INTERRUPTS-ENABLED* = %s\n", (read_TLS(INTERRUPTS_ENABLED, thread) == T) ? "T" : "NIL");
+    printf(" *GC-INHIBIT* = %s\n", read_TLS(GC_INHIBIT, thread) == LISP_T ? "T" : "NIL");
+    printf(" *GC-PENDING* = %s\n", read_TLS(GC_PENDING, thread) == LISP_T ? "T" : "NIL");
+    printf(" *INTERRUPTS-ENABLED* = %s\n",
+           read_TLS(INTERRUPTS_ENABLED, thread) == LISP_T ? "T" : "NIL");
 #ifdef STOP_FOR_GC_PENDING
-    printf(" *STOP-FOR-GC-PENDING* = %s\n", (read_TLS(STOP_FOR_GC_PENDING, thread) == T) ? "T" : "NIL");
+    printf(" *STOP-FOR-GC-PENDING* = %s\n",
+           read_TLS(STOP_FOR_GC_PENDING, thread) == LISP_T ? "T" : "NIL");
 #endif
     printf("Pending handler = %p\n", data->pending_handler);
 }
 
+static char* asm_routine_name(char* pc)
+{
+    struct code *c = (void*)asm_routines_start;
+    int offset = pc - code_text_start(c);
+    struct hash_table* ht = (void*)native_pointer(c->debug_info);
+    struct vector* v = (void*)native_pointer(ht->pairs);
+    int len = vector_len(v);
+    int i;
+    for (i = 2; i < len; i += 2) {
+        if (lowtag_of(v->data[i+1]) != LIST_POINTER_LOWTAG) continue;
+        struct cons* c = CONS(v->data[i+1]);
+        struct cons* cdr = CONS(c->cdr);
+        int from_byteindex = fixnum_value(c->car);
+        int to_byteindex = fixnum_value(cdr->car);
+        if (offset >= from_byteindex && offset <= to_byteindex) {
+            struct symbol* sym = SYMBOL(v->data[i]);
+            struct vector* string = VECTOR(decode_symbol_name(sym->name));
+            return (char*)string->data;
+        }
+    }
+    return "?";
+}
+
 static void print_backtrace_frame(char *pc, void *fp, int i, FILE *f) {
-    lispobj *p;
-    fprintf(f, "%4d: ", i);
-
-    p = component_ptr_from_pc(pc);
-
-    if (p) {
-        struct code *cp = (struct code *) p;
-        struct compiled_debug_fun *df = debug_function_from_pc(cp, pc);
+#ifdef BACKTRACE_SHOW_FRAME_SIZE
+    // This display is a little confusing.  It's the size of the frame that this
+    // frame will return to.
+    fprintf(f, "%4d: fp=%p [%5x] pc=%p ", i, fp, (int)(*(char**)fp-(char*)fp), pc);
+#else
+    fprintf(f, "%4d: fp=%p pc=%p ", i, fp, pc);
+#endif
+    struct code *code = (void*)component_ptr_from_pc(pc);
+    if (code) {
+        struct compiled_debug_fun *df = debug_function_from_pc(code, pc);
         if (df)
             print_entry_name(df->name, f);
+        else if (pc >= (char*)asm_routines_start && pc < (char*)asm_routines_end)
+            fprintf(f, "%s (asm)", asm_routine_name(pc));
         else
-            print_entry_points(cp, f);
-        fprintf(f, ", pc = %p, fp = %p", pc, fp);
+            fprintf(f, "{code_serialno=%x}", code_serialno(code));
+    } else if (gc_managed_heap_space_p((uword_t)pc)) {
+#ifdef LISP_FEATURE_X86
+        // can't actually have a PC inside a random object, it's got to be a frame
+        // that didn't set up the pointer chain, quite possibly a signal frame such as:
+        //   7: fp=0xd78c8460 pc=0xf7fb51b0 Foreign function __kernel_rt_sigreturn
+        //   8: fp=0xd78c8478 pc=0xd9c43159 (bad PC)
+        //   9: fp=0xd78c84ec pc=0xd849a17e (FLET SB-C::DO-1-USE :IN SB-C::TENSION-IF-IF-1)
+        // where, if you print the PC actually from the context, line 8 would be 0xd823ea78.
+        fprintf(f, "(bad PC)");
+#else
+        // It could be a generic-function with self-contained tramponline code,
+        // or the executable JMP instruction in an fdefn.
+        fprintf(f, "(unknown lisp object)");
+#endif
     } else {
 #ifdef LISP_FEATURE_OS_PROVIDES_DLADDR
         Dl_info info;
         if (dladdr(pc, &info)) {
-            fprintf(f, "Foreign function %s, pc = %p, fp = %p", info.dli_sname, pc, fp);
+            fprintf(f, "Foreign function %s", info.dli_sname);
         } else
 #endif
-            fprintf(f, "Foreign function, pc = %p, fp = %p", pc, fp);
+            fprintf(f, "Foreign function");
     }
 
     putc('\n', f);
@@ -520,7 +600,7 @@ static void print_backtrace_frame(char *pc, void *fp, int i, FILE *f) {
  * example when debugging threading deadlocks.
  */
 void NO_SANITIZE_MEMORY
-log_backtrace_from_fp(void *fp, int nframes, int start, FILE *f)
+log_backtrace_from_fp(struct thread* th, void *fp, int nframes, int start, FILE *f)
 {
   int i = start;
 
@@ -528,35 +608,31 @@ log_backtrace_from_fp(void *fp, int nframes, int start, FILE *f)
     void *ra;
     void *next_fp;
 
-    if (!x86_call_context(fp, &ra, &next_fp))
+    if (!x86_call_context(th, fp, &ra, &next_fp))
       break;
     print_backtrace_frame(ra, next_fp, i, f);
     fp = next_fp;
   }
 }
 void backtrace_from_fp(void *fp, int nframes, int start) {
-    log_backtrace_from_fp(fp, nframes, start, stdout);
+    log_backtrace_from_fp(get_sb_vm_thread(), fp, nframes, start, stdout);
 }
 
-void backtrace_from_context(os_context_t *context, int nframes) {
-#ifdef LISP_FEATURE_X86
-    void *fp = (void *)*os_context_register_addr(context,reg_EBP);
-#elif defined (LISP_FEATURE_X86_64)
-    void *fp = (void *)*os_context_register_addr(context,reg_RBP);
-#endif
-    print_backtrace_frame((void *)*os_context_pc_addr(context), fp, 0, stdout);
-    backtrace_from_fp(fp, nframes - 1, 1);
+void print_backtrace_from_context(os_context_t *context, int nframes, FILE* file) {
+    void *fp = (void *)os_context_frame_pointer(context);
+    print_backtrace_frame((void *)os_context_pc(context), fp, 0, file);
+    log_backtrace_from_fp(get_sb_vm_thread(), fp, nframes - 1, 1, file);
 }
 
 void
 lisp_backtrace(int nframes)
 {
-    struct thread *thread=arch_os_get_current_thread();
+    struct thread *thread = get_sb_vm_thread();
     int free_ici = fixnum_value(read_TLS(FREE_INTERRUPT_CONTEXT_INDEX,thread));
 
     if (free_ici) {
         os_context_t *context = nth_interrupt_context(free_ici - 1, thread);
-        backtrace_from_context(context, nframes);
+        print_backtrace_from_context(context, nframes, stdout);
     } else {
         void *fp;
 
@@ -583,10 +659,10 @@ int simple_fun_index_from_pc(struct code* code, char *pc)
     return -1;
 }
 
-static boolean __attribute__((unused)) print_lisp_fun_name(char* pc)
+static bool __attribute__((unused)) print_lisp_fun_name(char* pc)
 {
   struct code* code;
-  if (gc_managed_addr_p((uword_t)pc) &&
+  if (gc_managed_heap_space_p((uword_t)pc) &&
       (code = (void*)component_ptr_from_pc(pc)) != 0) {
       struct compiled_debug_fun* df = debug_function_from_pc(code, pc);
       if (df) {
@@ -599,25 +675,48 @@ static boolean __attribute__((unused)) print_lisp_fun_name(char* pc)
   return 0;
 }
 
-#ifdef LISP_FEATURE_LIBUNWIND_BACKTRACE
+#ifdef HAVE_LIBUNWIND
 #define UNW_LOCAL_ONLY
 #include <libunwind.h>
-#include "genesis/thread-instance.h"
+int get_sizeof_unw_context() { return sizeof (unw_context_t); }
+int get_sizeof_unw_cursor() { return sizeof (unw_cursor_t); }
+#ifdef LISP_FEATURE_DARWIN // slightly different libunwind. And it doesn't work for me
+int sb_unw_init(void* a, void* b) { return unw_init_local(a, b); }
+int sb_unw_get_pc(void* a, void* b) { return unw_get_reg(a, UNW_REG_IP, b); }
+#else
+int sb_unw_init(void* a, void* b) { return unw_init_local2(a, b, UNW_INIT_SIGNAL_FRAME); }
+int sb_unw_get_pc(void* a, void* b) { return unw_get_reg(a, UNW_TDEP_IP, b); }
+#endif
+int sb_unw_get_proc_name(void* a, void* b, int c, unw_word_t* d) { return unw_get_proc_name(a, b, c, d); }
+int sb_unw_step(void* a) { return unw_step(a); }
+#else
+int get_sizeof_unw_context() { return 0; }
+int get_sizeof_unw_cursor() { return 0; }
+int sb_unw_init(void* a, void* b) { lose("unw_init %p %p", a, b); }
+int sb_unw_get_proc_name(void* a, void* b, int c, void* d) { lose("unw_get_proc_name %p %p %d %p", a, b, c, d); }
+int sb_unw_step(void* a) { lose("unw_step %p", a); }
+int sb_unw_get_pc(void* a, void* b) { lose("unw_get_pc %p %p", a, b); }
+#endif
+
+#ifdef LISP_FEATURE_BACKTRACE_ON_SIGNAL
 #include "genesis/mutex.h"
 static __attribute__((unused))int backtrace_completion_pipe[2] = {-1,-1};
 void libunwind_backtrace(struct thread *th, os_context_t *context)
 {
-    char procname[100];
     fprintf(stderr, "Lisp thread @ %p, tid %d", th, (int)th->os_kernel_tid);
 #ifdef LISP_FEATURE_SB_THREAD
-    // the TLS area is not used if #-sb-thread. And if so, it must be "main thred"
+    // the TLS area is not used if #-sb-thread. And if so, it must be "main thread"
     struct thread_instance* lispthread = (void*)native_pointer(th->lisp_thread);
-    if (lispthread->name != NIL) {
+    if (lispthread->_name != NIL) {
         fprintf(stderr, " (\"");
-        print_string(VECTOR(lispthread->name), stderr);
+        print_string(VECTOR(lispthread->_name), stderr);
         fprintf(stderr, "\")");
     }
     putc('\n', stderr);
+    // In case you get no backtrace whatsoever, maybe at least see where the
+    // signal was received, probably in a function without the standard
+    // frame pointer setup.
+    fprintf(stderr, " interrupted @ PC %p\n", (void*)OS_CONTEXT_PC(context));
     if (lispthread->waiting_for != NIL) {
         fprintf(stderr, "waiting for %p", (void*)lispthread->waiting_for);
         if (instancep(lispthread->waiting_for)) {
@@ -636,6 +735,8 @@ void libunwind_backtrace(struct thread *th, os_context_t *context)
         putc('\n', stderr);
     }
 #endif
+#ifdef HAVE_LIBUNWIND
+    char procname[100];
     unw_cursor_t cursor;
     // "unw_init_local() is thread-safe as well as safe to use from a signal handler."
     // "unw_get_proc_name() is thread-safe. If cursor cp is in the local address-space,
@@ -659,29 +760,36 @@ void libunwind_backtrace(struct thread *th, os_context_t *context)
             fprintf(stderr, " %p ?\n", pc);
         }
     } while (unw_step(&cursor));
+#else
+    // If you don't have libunwind, this will almost surely not work,
+    // because we can't figure out how to get backwards past a signal frame.
+    log_backtrace_from_fp(th, (void*)*os_context_fp_addr(context), 100, 0, stderr);
+#endif
 }
 void backtrace_lisp_threads(int __attribute__((unused)) signal,
                                    siginfo_t __attribute__((unused)) *info,
                                    os_context_t *context)
 {
+    struct thread* this_thread = get_sb_vm_thread();
 #ifdef LISP_FEATURE_SB_THREAD
     if (backtrace_completion_pipe[1] >= 0) {
-        libunwind_backtrace(current_thread, context);
+        libunwind_backtrace(get_sb_vm_thread(), context);
         write(backtrace_completion_pipe[1], context /* any random byte */, 1);
         return;
     }
     struct thread *th;
     int nthreads = 0;
     for_each_thread(th) { ++nthreads; }
-    fprintf(stderr, "Caught backtrace-all signal in tid %d, %d threads\n",
-            (int)arch_os_get_current_thread()->os_kernel_tid, nthreads);
+    if (signal)
+        fprintf(stderr, "Caught backtrace-all signal in tid %d, %d threads\n",
+                (int)this_thread->os_kernel_tid, nthreads);
     // Would be nice if we could forcibly stop all the other threads,
     // but pthread_mutex_trylock is not safe to use in a signal handler.
     if (nthreads > 1) {
         pipe(backtrace_completion_pipe);
     }
     for_each_thread(th) {
-        if (th == arch_os_get_current_thread())
+        if (th == this_thread)
             libunwind_backtrace(th, context);
         else {
             char junk;
@@ -695,7 +803,56 @@ void backtrace_lisp_threads(int __attribute__((unused)) signal,
         backtrace_completion_pipe[0] = backtrace_completion_pipe[1] = -1;
     }
 #else
-    libunwind_backtrace(arch_os_get_current_thread(), context);
+    libunwind_backtrace(this_thread, context);
 #endif
 }
+static int watchdog_pipe[2] = {-1,-1};
+static pthread_t watchdog_tid;
+static void* watchdog_thread(void* arg) {
+    struct timeval timeout;
+    fd_set fds;
+    FD_ZERO(&fds);
+    for (;;) {
+        FD_SET(watchdog_pipe[0], &fds);
+        timeout.tv_sec = (long)arg;
+        timeout.tv_usec = 0;
+        int nfds = select(watchdog_pipe[0]+1, &fds, 0, 0, &timeout);
+        if (nfds < 1) {
+            // Ensure this message comes out in one piece even if nothing following it does.
+            char msg[] = "Watchdog timer expired\n"; write(2, msg, sizeof msg-1);
+            backtrace_lisp_threads(0, 0, 0);
+            _exit(1); // cause the test suite to exit with failure
+        }
+        if (!FD_ISSET(watchdog_pipe[0], &fds))
+            lose("Watchdog got strange file descriptor mask");
+        char byte;
+        int n = read(watchdog_pipe[0], &byte, 1);
+        if (n != 1) lose("watchdog could not read a byte?");
+        fprintf(stderr, "watchdog got byte %d\n", byte);
+        switch (byte) {
+        case 0: return 0;
+        case 1: break; // restart the timer
+        default: lose("watchdog: bad byte");
+        }
+    }
+    return 0;
+}
+void start_sbcl_watchdog(int sec) {
+    if (pipe(watchdog_pipe)) lose("Can't make watchdog pipe");
+    pthread_create(&watchdog_tid, 0, watchdog_thread, (void*)(long)sec);
+    char msg[] = "Started watchdog thread\n"; write(2, msg, sizeof msg-1);
+}
+void reset_sbcl_watchdog_timer() {
+    char c[1] = {1};
+    write(watchdog_pipe[1], c, 1);
+}
+void stop_sbcl_watchdog() {
+    char c[1] = {0};
+    write(watchdog_pipe[1], c, 1);
+    close(watchdog_pipe[1]);
+    void* result;
+    pthread_join(watchdog_tid, &result);
+    close(watchdog_pipe[0]);
+    watchdog_pipe[0] = watchdog_pipe[1] = -1;
+}
 #endif
